# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.keras.backend as K
from numpy.testing import assert_allclose, assert_almost_equal

from decomon.core import BoxDomain, ForwardMode
from decomon.layers.utils import add, max_, maximum, minus, relu_
from decomon.utils import subtract


def test_get_upper_multi_box(odd, floatx, decimal, helpers):
    inputs = helpers.get_tensor_decomposition_multid_box(odd)
    inputs_ = helpers.get_standard_values_multid_box(odd)

    x, y, x_0, u_c, W_u, b_u, _, _, _, _, _ = inputs
    x_, y_, x_0_, u_c_, W_u_, b_u_, _, _, _, _, _ = inputs_

    # compute maximum
    x_min_ = x_0_[:, 0][:, :, None]
    x_max_ = x_0_[:, 1][:, :, None]
    upper_pred = np.sum(np.minimum(W_u_, 0) * x_min_ + np.maximum(W_u_, 0) * x_max_, 1) + b_u_

    upper = BoxDomain().get_upper(x_0, W_u, b_u)

    f_upper = K.function([x_0, W_u, b_u], upper)

    upper_ = f_upper([x_0_, W_u_, b_u_])

    assert_allclose(upper_pred, upper_)


def test_get_upper_box_numpy(n, floatx, decimal, helpers):
    inputs = helpers.get_tensor_decomposition_1d_box()
    inputs_ = helpers.get_standard_values_1d_box(n)

    x, y, x_0, u_c, W_u, b_u, _, _, _, _, _ = inputs
    x_, y_, x_0_, u_c_, W_u_, b_u_, _, _, _, _, _ = inputs_

    x_expand = x_ + np.zeros_like(x_)
    n_expand = len(W_u_.shape) - len(x_expand.shape)
    for i in range(n_expand):
        x_expand = np.expand_dims(x_expand, -1)

    upper_pred = np.sum(W_u_ * x_expand, 1) + b_u_
    upper_pred = upper_pred.max(0)

    upper = BoxDomain().get_upper(x_0, W_u, b_u)

    f_upper = K.function([x_0, W_u, b_u], upper)
    upper_ = f_upper([x_0_, W_u_, b_u_]).max()

    assert_allclose(upper_pred, upper_)


def test_get_upper_box(n, floatx, decimal, helpers):
    inputs = helpers.get_tensor_decomposition_1d_box()
    inputs_ = helpers.get_standard_values_1d_box(n)

    x, y, x_0, u_c, W_u, b_u, _, _, _, _, _ = inputs
    _, _, x_0_, u_c_, W_u_, b_u_, _, _, _, _, _ = inputs_

    upper = BoxDomain().get_upper(x_0, W_u, b_u)

    f_upper = K.function([x_0, W_u, b_u], upper)

    f_u = K.function(inputs, u_c)

    output = inputs_[1]

    upper_ = f_upper([x_0_, W_u_, b_u_])
    u_c_ = f_u(inputs_)

    assert_almost_equal(
        np.clip(output - upper_, 0, np.inf),
        np.zeros_like(output),
        decimal=decimal,
        err_msg="upper_<y_ in call {}".format(n),
    )

    assert_almost_equal(
        np.clip(u_c_ - upper_, 0, np.inf),
        np.zeros_like(output),
        decimal=decimal,
        err_msg="upper_<y_ in call {}".format(n),
    )


def test_get_lower_box(n, floatx, decimal, helpers):
    inputs = helpers.get_tensor_decomposition_1d_box()
    inputs_ = helpers.get_standard_values_1d_box(n)
    x, y, x_0, _, _, _, l_c, W_l, b_l, _, _ = inputs

    lower = BoxDomain().get_lower(x_0, W_l, b_l)

    f_lower = K.function(inputs, lower)
    f_l = K.function(inputs, l_c)

    input_ref = helpers.get_input_ref_from_full_inputs(inputs_)

    lower_ = f_lower(inputs_)
    l_c_ = f_l(inputs_)

    assert_almost_equal(
        np.clip(lower_ - input_ref, 0, np.inf),
        np.zeros_like(input_ref),
        decimal=decimal,
        err_msg="lower_> y_ in call {}".format(n),
    )
    assert_almost_equal(
        np.clip(lower_ - l_c_, 0, np.inf),
        np.zeros_like(input_ref),
        decimal=decimal,
        err_msg="lower_> y_ in call {}".format(n),
    )


def test_get_lower_upper_box(n, floatx, decimal, helpers):
    inputs = helpers.get_tensor_decomposition_1d_box()
    inputs_ = helpers.get_standard_values_1d_box(n)
    x, y, x_0, _, W_u, b_u, _, W_l, b_l, _, _ = inputs

    lower = BoxDomain().get_lower(x_0, W_l, b_l)
    upper = BoxDomain().get_upper(x_0, W_u, b_u)

    f_lower = K.function(inputs, lower)
    f_upper = K.function(inputs, upper)

    lower_ = f_lower(inputs_)
    upper_ = f_upper(inputs_)

    assert_almost_equal(
        np.clip(lower_ - upper_, 0, np.inf),
        np.zeros_like(lower_),
        decimal=decimal,
        err_msg="lower_> upper_ in call {}".format(n),
    )


def test_relu_1D_box(n, mode, floatx, decimal, helpers):
    dc_decomp = True

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # original output
    f_ref = K.function(inputs, K.relu(input_ref))
    output_ref_ = f_ref(inputs_)

    # decomon output
    output = relu_(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    f_decomon = K.function(inputs, output)
    outputs_ = f_decomon(inputs_)

    # full outputs
    (
        z_output,
        u_c_output,
        w_u_output,
        b_u_output,
        l_c_output,
        w_l_output,
        b_l_output,
        h_output,
        g_output,
    ) = helpers.get_full_outputs_from_outputs_for_mode(
        outputs_for_mode=outputs_,
        full_inputs=inputs_,
        mode=mode,
        dc_decomp=dc_decomp,
    )

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )

    # Further checks
    (
        x_tensor,
        y_tensor,
        z_tensor,
        u_c_tensor,
        W_u_tensor,
        b_u_tensor,
        l_c_tensor,
        W_l_tensor,
        b_l_tensor,
        h_tensor,
        g_tensor,
    ) = inputs  # tensors
    (
        x_input,
        y_input,
        z_input,
        u_c_input,
        W_u_input,
        b_u_input,
        l_c_input,
        W_l_input,
        b_l_input,
        h_input,
        g_input,
    ) = inputs_  # numpy values

    # lower and upper bounds from affine coefficients
    lower = BoxDomain().get_lower(z_tensor, W_l_tensor, b_l_tensor)
    upper = BoxDomain().get_upper(z_tensor, W_u_tensor, b_u_tensor)
    f_lower = K.function(inputs, lower)
    f_upper = K.function(inputs, upper)
    lower_ = np.min(f_lower(inputs_))
    upper_ = np.max(f_upper(inputs_))

    if upper_ <= 0:
        # check that we find this case !
        if mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
            assert_almost_equal(
                w_u_output,
                np.zeros_like(w_u_output),
                decimal=decimal,
                err_msg="w_u_!=0 but upper_<=0 in call {}".format(n),
            )
            assert_almost_equal(
                b_u_output,
                np.zeros_like(b_u_output),
                decimal=decimal,
                err_msg="b_u_!=0 but upper_<=0 in call {}".format(n),
            )
            assert_almost_equal(
                w_l_output,
                np.zeros_like(w_l_output),
                decimal=decimal,
                err_msg="w_l_!=0 but upper_<=0 in call {}".format(n),
            )
            assert_almost_equal(
                b_l_output,
                np.zeros_like(b_l_output),
                decimal=decimal,
                err_msg="b_l_!=0 but upper_<=0 in call {}".format(n),
            )

        if mode in [ForwardMode.IBP, ForwardMode.HYBRID]:
            assert_almost_equal(
                u_c_output,
                np.zeros_like(u_c_output),
                decimal=decimal,
                err_msg="u_c_!=0 but upper_<=0 in call {}".format(n),
            )
            assert_almost_equal(
                l_c_output,
                np.zeros_like(l_c_output),
                decimal=decimal,
                err_msg="l_c_!=0 but upper_<=0 in call {}".format(n),
            )

    if lower_ >= 0:
        # check that we find this case !
        if mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
            assert_almost_equal(
                w_u_output,
                W_u_input,
                decimal=decimal,
                err_msg="w_u_!=W_u but lower_>=0 in call {}".format(n),
            )
            assert_almost_equal(
                b_u_output,
                b_u_input,
                decimal=decimal,
                err_msg="b_u_!=b_u but lower_>=0  in call {}".format(n),
            )
            assert_almost_equal(
                w_l_output,
                W_l_input,
                decimal=decimal,
                err_msg="w_l_!=W_l but lower_>=0 upper_<=0 in call {}".format(n),
            )
            assert_almost_equal(
                b_l_output,
                b_l_input,
                decimal=decimal,
                err_msg="b_l_!=b_l but lower_>=0 upper_<=0 in call {}".format(n),
            )
        if mode in [ForwardMode.IBP, ForwardMode.HYBRID]:
            assert_almost_equal(
                u_c_output,
                u_c_input,
                decimal=decimal,
                err_msg="u_c_!=u_c but lower_>=0  in call {}".format(n),
            )
            assert_almost_equal(
                l_c_output,
                l_c_input,
                decimal=decimal,
                err_msg="l_c_!=l_c but lower_>=0  in call {}".format(n),
            )

    if mode != ForwardMode.IBP:
        assert_almost_equal(
            z_input[:, 0], z_output[:, 0], decimal=decimal, err_msg="the lower bound should be unchanged"
        )
        assert_almost_equal(
            z_input[:, 1], z_output[:, 1], decimal=decimal, err_msg="the upper bound should be unchanged"
        )


def add_op(x, y):
    return x + y


def subtract_op(x, y):
    return x - y


def minus_op(x):
    return -x


@pytest.mark.parametrize(
    "decomon_func, tensor_func",
    [
        (add, add_op),
        (subtract, subtract_op),
        (maximum, K.maximum),
    ],
)
def test_func_with_2_inputs(decomon_func, tensor_func, odd, mode, floatx, decimal, helpers):
    dc_decomp = True

    #  tensor inputs
    inputs_0 = helpers.get_tensor_decomposition_multid_box(odd)
    inputs_1 = helpers.get_tensor_decomposition_multid_box(odd)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)
    input_ref_0 = helpers.get_input_ref_from_full_inputs(inputs=inputs_0)
    input_ref_1 = helpers.get_input_ref_from_full_inputs(inputs=inputs_1)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd)

    # original output
    f_ref = K.function(inputs_0 + inputs_1, tensor_func(input_ref_0, input_ref_1))
    output_ref_ = f_ref(inputs_ + inputs_)

    # decomon output
    output = decomon_func(inputs_for_mode_0, inputs_for_mode_1, dc_decomp=dc_decomp, mode=mode)
    f_decomon = K.function(inputs_0 + inputs_1, output)
    outputs_ = f_decomon(inputs_ + inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "decomon_func, tensor_func, tensor_func_kwargs",
    [
        (minus, minus_op, None),
        (max_, K.max, {"axis": -1}),
    ],
)
def test_func_with_1_input(decomon_func, tensor_func, tensor_func_kwargs, odd, mode, floatx, decimal, helpers):
    dc_decomp = True
    if tensor_func_kwargs is None:
        tensor_func_kwargs = {}

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # original output
    f_ref = K.function(inputs, tensor_func(input_ref, **tensor_func_kwargs))
    output_ref_ = f_ref(inputs_)

    # decomon output
    output = decomon_func(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    f_decomon = K.function(inputs, output)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


# DC_DECOMP = FALSE
def test_max_nodc(odd, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID

    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    output = max_(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    f_max = K.function(inputs, output)
    assert len(f_max(inputs_)) == 7


def test_maximum_nodc(odd, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID

    inputs_0 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)
    input_ref_0 = helpers.get_input_ref_from_full_inputs(inputs=inputs_0)
    input_ref_1 = helpers.get_input_ref_from_full_inputs(inputs=inputs_1)

    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)
    output = maximum(inputs_for_mode_0, inputs_for_mode_1, dc_decomp=dc_decomp, mode=mode)

    f_ref = K.function(inputs_0 + inputs_1, K.maximum(input_ref_0, input_ref_1))
    f_maximum = K.function(inputs_0 + inputs_1, output)

    assert len(f_maximum(inputs_ + inputs_)) == 7
    f_ref(inputs_ + inputs_)


def test_minus_nodc(odd, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID

    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)
    output = minus(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)

    f_ref = K.function(inputs, -inputs[1])
    f_minus = K.function(inputs, output)

    assert len(f_minus(inputs_)) == 7
    f_ref(inputs_)


def test_add_nodc(odd, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID

    inputs_0 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)
    input_ref_0 = helpers.get_input_ref_from_full_inputs(inputs=inputs_0)
    input_ref_1 = helpers.get_input_ref_from_full_inputs(inputs=inputs_1)

    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    output = add(inputs_for_mode_0, inputs_for_mode_1, dc_decomp=dc_decomp, mode=mode)
    f_ref = K.function(inputs_0 + inputs_1, input_ref_0 + input_ref_1)
    f_add = K.function(inputs_0 + inputs_1, output)
    assert len(f_add(inputs_ + inputs_)) == 7
    f_ref(inputs_ + inputs_)


def test_relu_1D_box_nodc(n, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID

    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs

    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    output = relu_(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    lower = BoxDomain().get_lower(z, W_l, b_l)
    upper = BoxDomain().get_upper(z, W_u, b_u)

    f_lower = K.function(inputs, lower)
    f_upper = K.function(inputs, upper)

    np.min(f_lower(inputs_))
    np.max(f_upper(inputs_))

    f_relu_ = K.function(inputs, output)
    assert len(f_relu_(inputs_)) == 7



from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Input, InputLayer, Lambda, Layer
from tensorflow.keras.models import Model

from decomon.backward_layers.core import BackwardLayer
from decomon.core import BoxDomain, PerturbationDomain, Slope, get_mode
from decomon.layers.convert import to_decomon
from decomon.models.backward_cloning import convert_backward
from decomon.models.forward_cloning import (
    LayerMapDict,
    OutputMapDict,
    convert_forward,
    convert_forward_functional_model,
)
from decomon.models.models import DecomonModel
from decomon.models.utils import (
    Convert2Mode,
    ConvertMethod,
    FeedDirection,
    convert_deellip_to_keras,
    get_direction,
    get_ibp_affine_from_method,
    get_input_tensors,
    preprocess_layer,
    split_activation,
)


def _clone_keras_model(model: Model, layer_fn: Callable[[Layer], List[Layer]]) -> Model:
    if model.inputs is None:
        raise ValueError("model.inputs must be not None. You should call the model on a batch of data.")

    # initialize output_map and layer_map to avoid
    #   - recreating input layers
    #   - and converting input layers and have a cycle around them
    output_map: OutputMapDict = {id(input_tensor.node): [input_tensor] for input_tensor in model.inputs}
    layer_map: LayerMapDict = {id(input_tensor.node): input_tensor.node.outbound_layer for input_tensor in model.inputs}
    _, output, _, _ = convert_forward_functional_model(
        model=model,
        input_tensors=model.inputs,
        softmax_to_linear=False,
        layer_fn=layer_fn,
        output_map=output_map,
        layer_map=layer_map,
    )

    return Model(
        inputs=model.inputs,
        outputs=output,
    )


def split_activations_in_keras_model(
    model: Model,
) -> Model:
    return _clone_keras_model(model=model, layer_fn=split_activation)


def convert_deellip_layers_in_keras_model(
    model: Model,
) -> Model:
    return _clone_keras_model(model=model, layer_fn=_convert_deellip_to_keras)


def _convert_deellip_to_keras(layer: Layer) -> List[Layer]:
    return [convert_deellip_to_keras(layer=layer)]


def preprocess_keras_model(
    model: Model,
) -> Model:
    return _clone_keras_model(model=model, layer_fn=preprocess_layer)


# create status
def convert(
    model: Model,
    input_tensors: List[tf.Tensor],
    method: Union[str, ConvertMethod] = ConvertMethod.CROWN,
    ibp: bool = False,
    affine: bool = False,
    back_bounds: Optional[List[tf.Tensor]] = None,
    layer_fn: Callable[..., Layer] = to_decomon,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    input_dim: int = -1,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    forward_map: Optional[OutputMapDict] = None,
    shared: bool = True,
    softmax_to_linear: bool = True,
    finetune_forward: bool = False,
    finetune_backward: bool = False,
    final_ibp: bool = False,
    final_affine: bool = False,
    **kwargs: Any,
) -> Tuple[List[tf.Tensor], List[tf.Tensor], Union[LayerMapDict, Dict[int, BackwardLayer]], Optional[OutputMapDict],]:
    if back_bounds is None:
        back_bounds = []
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if finetune:
        finetune_forward = True
        finetune_backward = True

    if isinstance(method, str):
        method = ConvertMethod(method.lower())

    # prepare the Keras Model: split non-linear activation functions into separate Activation layers
    model = preprocess_keras_model(model)

    layer_map: Union[LayerMapDict, Dict[int, BackwardLayer]]

    if method != ConvertMethod.CROWN:
        input_tensors, output, layer_map, forward_map = convert_forward(
            model=model,
            input_tensors=input_tensors,
            layer_fn=layer_fn,
            slope=slope,
            input_dim=input_dim,
            dc_decomp=False,
            perturbation_domain=perturbation_domain,
            ibp=ibp,
            affine=affine,
            finetune=finetune_forward,
            shared=shared,
            softmax_to_linear=softmax_to_linear,
            back_bounds=back_bounds,
        )

    if get_direction(method) == FeedDirection.BACKWARD:
        input_tensors, output, layer_map, forward_map = convert_backward(
            model=model,
            input_tensors=input_tensors,
            back_bounds=back_bounds,
            slope=slope,
            perturbation_domain=perturbation_domain,
            ibp=ibp,
            affine=affine,
            finetune=finetune_backward,
            forward_map=forward_map,
            final_ibp=final_ibp,
            final_affine=final_affine,
            **kwargs,
        )
    else:
        # check final_ibp and final_affine
        mode_from = get_mode(ibp, affine)
        mode_to = get_mode(final_ibp, final_affine)
        output = Convert2Mode(mode_from=mode_from, mode_to=mode_to, perturbation_domain=perturbation_domain)(output)

    # build decomon model
    return input_tensors, output, layer_map, forward_map


def clone(
    model: Model,
    layer_fn: Callable[..., Layer] = to_decomon,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    perturbation_domain: Optional[PerturbationDomain] = None,
    method: Union[str, ConvertMethod] = ConvertMethod.CROWN,
    back_bounds: Optional[List[tf.Tensor]] = None,
    finetune: bool = False,
    shared: bool = True,
    finetune_forward: bool = False,
    finetune_backward: bool = False,
    extra_inputs: Optional[List[tf.Tensor]] = None,
    to_keras: bool = True,
    final_ibp: Optional[bool] = None,
    final_affine: Optional[bool] = None,
    **kwargs: Any,
) -> DecomonModel:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if back_bounds is None:
        back_bounds = []
    if extra_inputs is None:
        extra_inputs = []
    if not isinstance(model, Model):
        raise ValueError("Expected `model` argument " "to be a `Model` instance, got ", model)

    ibp, affine = get_ibp_affine_from_method(method)
    if final_ibp is None:
        final_ibp = ibp
    if final_affine is None:
        final_affine = affine

    if isinstance(method, str):
        method = ConvertMethod(method.lower())

    if not to_keras:
        raise NotImplementedError("Only convert to Keras for now.")

    if finetune:
        finetune_forward = True
        finetune_backward = True

    z_tensor, input_tensors = get_input_tensors(
        model=model,
        perturbation_domain=perturbation_domain,
        ibp=ibp,
        affine=affine,
    )

    _, output, _, _ = convert(
        model,
        layer_fn=layer_fn,
        slope=slope,
        input_tensors=input_tensors,
        back_bounds=back_bounds,
        method=method,
        ibp=ibp,
        affine=affine,
        input_dim=-1,
        perturbation_domain=perturbation_domain,
        finetune=finetune,
        shared=shared,
        softmax_to_linear=True,
        layer_map={},
        forward_map={},
        finetune_forward=finetune_forward,
        finetune_backward=finetune_backward,
        final_ibp=final_ibp,
        final_affine=final_affine,
    )

    back_bounds_from_inputs = [elem for elem in back_bounds if isinstance(elem._keras_history.layer, InputLayer)]

    return DecomonModel(
        inputs=[z_tensor] + back_bounds_from_inputs + extra_inputs,
        outputs=output,
        perturbation_domain=perturbation_domain,
        dc_decomp=False,
        method=method,
        ibp=final_ibp,
        affine=final_affine,
        finetune=finetune,
        shared=shared,
        backward_bounds=(len(back_bounds) > 0),
    )



# Goal: compute bounds with lirpa and decomon and assess that they are the same (up to numerical precision)

import numpy as np
import onnx
import torchvision
from auto_LiRPA import BoundedModule, BoundedTensor
from auto_LiRPA.perturbations import PerturbationLpNorm
from numpy.testing import assert_almost_equal
from onnx2keras import onnx_to_keras
from onnx2torch import convert

from decomon.models.convert import clone
from decomon.models.utils import ConvertMethod


@pytest.mark.parametrize(
    "path, onnx_filename, method, eps, N",
    [
        (".", "mnist_relu_3_50_comparison.onnx", "IBP", 0.01, 3),
        (".", "mnist_relu_3_50_comparison.onnx", "IBP", 0.1, 3),
        (".", "mnist_relu_3_50_comparison.onnx", "IBP+backward", 0.01, 3),
        (".", "mnist_relu_3_50_comparison.onnx", "IBP+backward", 0.1, 3),
        (".", "mnist_relu_3_50_comparison.onnx", "backward", 0.01, 3),
        (".", "mnist_relu_3_50_comparison.onnx", "backward", 0.1, 3),
    ],
)
def test_with_lirpa(path, onnx_filename, method, eps, N):
    decimal = 4
    equ_method = {
        "IBP": ConvertMethod.FORWARD_IBP,
        "backward": ConvertMethod.CROWN,
        "IBP+backward": ConvertMethod.CROWN_FORWARD_IBP,
    }

    # use use onnx2torch to load the network in torch
    onnx_model = onnx.load(onnx_filename)
    torch_model = clone(onnx_model)

    ## Step 2: Prepare dataset as usual
    test_data = torchvision.datasets.MNIST(
        "./data", train=False, download=True, transform=torchvision.transforms.ToTensor()
    )
    N = 3
    # we consider the first N test examples correctly classified by the network
    image = test_data.data[: 2 * N].view(2 * N, 784) / 255.0 + 0.0
    # apply lirpa bounds
    norm = np.inf

    ptb = PerturbationLpNorm(norm=norm, eps=eps)
    image_tensor = BoundedTensor(image, ptb)
    lirpa_model = BoundedModule(torch_model, image, device="cpu")

    lb_, ub_ = lirpa_model.compute_bounds(x=(image_tensor,), method=method)
    lb_p = lb_.detach().numpy()
    ub_p = ub_.detach().numpy()

    onnx_model = onnx.load(onnx_filename)
    k_model = onnx_to_keras(onnx_model, ["0"])

    decomon_model = clone(k_model, method=equ_method[method], final_ibp=True, final_affine=False)

    image_ = image.detach().numpy()
    box = np.concatenate([image_[:, None] - eps, image_[:, None] + eps], 1)
    ub_k, lb_k = decomon_model.predict(box)

    assert_almost_equal(ub_k, ub_p, decimal=decimal, err_msg="error")
    assert_almost_equal(lb_k, lb_p, decimal=decimal, err_msg="error")



import tensorflow.keras.backend as K
from tensorflow.keras.layers import MaxPooling2D

from decomon.layers.maxpooling import DecomonMaxPooling2D


def test_MaxPooling2D_box(mode, floatx, decimal, helpers):
    odd, m_0, m_1 = 0, 0, 1
    data_format = "channels_last"
    dc_decomp = True
    fast = False
    kwargs_layer = dict(pool_size=(2, 2), strides=(2, 2), padding="valid", dtype=K.floatx())

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)

    # keras & decomon layer
    keras_layer = MaxPooling2D(**kwargs_layer)
    decomon_layer = DecomonMaxPooling2D(dc_decomp=dc_decomp, fast=fast, mode=mode, **kwargs_layer)

    # original output
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)
    output_ref_ = f_ref(inputs_)

    # decomon outputs
    outputs = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        dc_decomp=dc_decomp,
        decimal=decimal,
        mode=mode,
    )



# creating toy network and assess that the decomposition is correct


import numpy as np
import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Activation, Dense, Flatten, Input, Reshape
from tensorflow.keras.models import Sequential

from decomon.core import ForwardMode, Slope, get_affine, get_ibp, get_mode
from decomon.layers.decomon_reshape import DecomonReshape
from decomon.models import clone
from decomon.models.convert import FeedDirection, get_direction
from decomon.models.utils import (
    ConvertMethod,
    get_ibp_affine_from_method,
    has_merge_layers,
)

try:
    import deel.lip
except ImportError:
    deel_lip_available = False
else:
    deel_lip_available = True
    from deel.lip.activations import GroupSort
    from deel.lip.layers import (
        FrobeniusDense,
        ScaledL2NormPooling2D,
        SpectralConv2D,
        SpectralDense,
    )
    from deel.lip.model import Sequential as DeellipSequential


deel_lip_skip_reason = "deel-lip is not available"


def test_convert_1D(n, method, mode, floatx, decimal, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    dc_decomp = False
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)
    input_decomon_ = helpers.get_inputs_np_for_decomon_model_from_full_inputs(inputs=inputs_)

    #  keras model and output of reference
    ref_nn = helpers.toy_network_tutorial(dtype=K.floatx())
    output_ref_ = ref_nn.predict(input_ref_)

    # decomon conversion
    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine)

    #  decomon outputs
    outputs_ = decomon_model.predict(input_decomon_)

    #  check bounds consistency
    helpers.assert_decomon_model_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


def test_convert_1D_forward_slope(slope, helpers):
    ibp = True
    affine = True
    dc_decomp = False

    n, method, mode = 0, "forward-hybrid", "hybrid"
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    ref_nn = helpers.toy_network_tutorial(dtype=K.floatx())
    ref_nn(input_ref)

    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine, slope=slope)

    # check slope of activation layers
    for layer in decomon_model.layers:
        if layer.__class__.__name__.endswith("Activation"):
            assert layer.slope == Slope(slope)


def test_convert_1D_backward_slope(slope, helpers):
    n, method, mode = 0, "crown-forward-hybrid", "hybrid"
    ibp = True
    affine = True
    dc_decomp = False

    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    ref_nn = helpers.toy_network_tutorial(dtype=K.floatx())
    ref_nn(input_ref)

    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine, slope=slope)

    # check slope of layers with activation
    for layer in decomon_model.layers:
        layer_class_name = layer.__class__.__name__
        if layer_class_name.endswith("Activation"):
            assert layer.slope == Slope(slope)


def test_name_forward():
    layers = []
    layers.append(Dense(1, input_dim=1))
    layers.append(Dense(1, name="superman"))  # specify the dimension of the input space
    layers.append(Activation("relu"))
    layers.append(Dense(1, activation="relu", name="batman"))
    model = Sequential(layers)

    decomon_model_f = clone(model=model, method=ConvertMethod.FORWARD_HYBRID)
    nb_superman_layers = len([layer for layer in decomon_model_f.layers if layer.name.startswith("superman_")])
    assert nb_superman_layers == 1
    nb_batman_layers = len([layer for layer in decomon_model_f.layers if layer.name.startswith("batman_")])
    assert nb_batman_layers == 2


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_name_forward_deellip():
    layers = []
    layers.append(Dense(1, input_dim=1))
    layers.append(SpectralDense(1, name="superman"))  # specify the dimension of the input space
    layers.append(Activation("relu"))
    layers.append(SpectralDense(1, activation=GroupSort(n=1), name="batman"))
    model = Sequential(layers)

    decomon_model_f = clone(model=model, method=ConvertMethod.FORWARD_HYBRID)
    nb_superman_layers = len([layer for layer in decomon_model_f.layers if layer.name.startswith("superman_")])
    assert nb_superman_layers == 1
    nb_batman_layers = len([layer for layer in decomon_model_f.layers if layer.name.startswith("batman_")])
    assert nb_batman_layers == 2


def test_name_backward():
    layers = []
    layers.append(Dense(1, input_dim=1))
    layers.append(Dense(1, name="superman"))  # specify the dimension of the input space
    layers.append(Activation("relu"))
    layers.append(Dense(1, activation="relu", name="batman"))
    model = Sequential(layers)

    decomon_model_b = clone(model=model, method=ConvertMethod.CROWN_FORWARD_HYBRID)
    nb_superman_layers = len([layer for layer in decomon_model_b.layers if layer.name.startswith("superman_")])
    assert nb_superman_layers == 2
    nb_batman_layers = len([layer for layer in decomon_model_b.layers if layer.name.startswith("batman_")])
    assert nb_batman_layers == 3


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_name_backward_deellip():
    layers = []
    layers.append(Dense(1, input_dim=1))
    layers.append(SpectralDense(1, name="superman"))  # specify the dimension of the input space
    layers.append(Activation("relu"))
    layers.append(SpectralDense(1, activation="relu", name="batman"))
    model = Sequential(layers)

    decomon_model_b = clone(model=model, method=ConvertMethod.CROWN_FORWARD_HYBRID)
    nb_superman_layers = len([layer for layer in decomon_model_b.layers if layer.name.startswith("superman_")])
    assert nb_superman_layers == 2
    nb_batman_layers = len([layer for layer in decomon_model_b.layers if layer.name.startswith("batman_")])
    assert nb_batman_layers == 3


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_clone_full_deellip_model_forward(method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    if get_direction(method) == FeedDirection.BACKWARD:
        # skip as BackwardConv2D not yet ready
        pytest.skip(f"BackwardConv2D not yet fully implemented")

    decimal = 4
    data_format = "channels_last"
    odd, m_0, m_1 = 0, 0, 1
    dc_decomp = False
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)
    input_ref_min_, input_ref_max_ = helpers.get_input_ref_bounds_from_full_inputs(inputs_)

    # flatten inputs
    preprocess_layer = Flatten(data_format=data_format)
    input_ref_reshaped_ = preprocess_layer(input_ref_).numpy()
    input_ref_min_reshaped_ = preprocess_layer(input_ref_min_).numpy()
    input_ref_max_reshaped_ = preprocess_layer(input_ref_max_).numpy()

    # decomon inputs
    input_decomon_ = np.concatenate((input_ref_min_reshaped_[:, None], input_ref_max_reshaped_[:, None]), axis=1)

    # deel-lip model and output of reference
    image_data_shape = input_ref_.shape[1:]  # image shape: before flattening
    flatten_input_shape = (np.prod(image_data_shape),)
    ref_nn = DeellipSequential(
        [
            Reshape(target_shape=image_data_shape, input_shape=flatten_input_shape),
            # Lipschitz layers preserve the API of their superclass ( here Conv2D )
            # an optional param is available: k_coef_lip which control the lipschitz
            # constant of the layer
            SpectralConv2D(
                filters=16,
                kernel_size=(3, 3),
                activation=GroupSort(2),
                use_bias=True,
                kernel_initializer="orthogonal",
            ),
            # our layers are fully interoperable with existing keras layers
            Flatten(),
            SpectralDense(
                32,
                activation=GroupSort(2),
                use_bias=True,
                kernel_initializer="orthogonal",
            ),
            FrobeniusDense(10, activation=None, use_bias=False, kernel_initializer="orthogonal"),
        ],
        # similary model has a parameter to set the lipschitz constant
        # to set automatically the constant of each layer
        k_coef_lip=1.0,
        name="hkr_model",
    )
    output_ref_ = ref_nn.predict(input_ref_reshaped_)

    # decomon conversion
    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine)

    #  decomon outputs
    outputs_ = decomon_model.predict(input_decomon_)

    #  check bounds consistency
    z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = helpers.get_full_outputs_from_outputs_for_mode(
        outputs_for_mode=outputs_, mode=mode, dc_decomp=dc_decomp, full_inputs=inputs_
    )
    helpers.assert_output_properties_box(
        input_ref_reshaped_,
        output_ref_,
        h_,
        g_,
        input_ref_min_reshaped_,
        input_ref_max_reshaped_,
        u_c_,
        w_u_,
        b_u_,
        l_c_,
        w_l_,
        b_l_,
        decimal=decimal,
    )


def test_convert_toy_models_1d(toy_model_1d, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    decimal = 4
    n = 6
    dc_decomp = False
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)
    input_decomon_ = helpers.get_inputs_np_for_decomon_model_from_full_inputs(inputs=inputs_)

    #  keras model and output of reference
    ref_nn = toy_model_1d
    output_ref_ = ref_nn.predict(input_ref_)

    # decomon conversion
    if (get_direction(method) == FeedDirection.BACKWARD) and has_merge_layers(ref_nn):
        # skip models with merge layers in backward direction as not yet implemented
        with pytest.raises(NotImplementedError):
            decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine)
        return

    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine)

    #  decomon outputs
    outputs_ = decomon_model.predict(input_decomon_)

    #  check bounds consistency
    helpers.assert_decomon_model_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


def test_convert_cnn(method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    if get_direction(method) == FeedDirection.BACKWARD:
        # skip as BackwardConv2D not yet ready
        pytest.skip(f"BackwardConv2D not yet fully implemented")

    decimal = 4
    data_format = "channels_last"
    odd, m_0, m_1 = 0, 0, 1

    dc_decomp = False
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)
    input_ref_min_, input_ref_max_ = helpers.get_input_ref_bounds_from_full_inputs(inputs_)

    # flatten inputs
    preprocess_layer = Flatten(data_format=data_format)
    input_ref_reshaped_ = preprocess_layer(input_ref_).numpy()
    input_ref_min_reshaped_ = preprocess_layer(input_ref_min_).numpy()
    input_ref_max_reshaped_ = preprocess_layer(input_ref_max_).numpy()

    input_decomon_ = np.concatenate((input_ref_min_reshaped_[:, None], input_ref_max_reshaped_[:, None]), axis=1)

    #  keras model and output of reference
    image_data_shape = input_ref_.shape[1:]  # image shape: before flattening
    ref_nn = helpers.toy_struct_cnn(dtype=K.floatx(), image_data_shape=image_data_shape)
    output_ref_ = ref_nn.predict(input_ref_reshaped_)

    # decomon conversion
    decomon_model = clone(ref_nn, method=method, final_ibp=ibp, final_affine=affine)

    #  decomon outputs
    outputs_ = decomon_model.predict(input_decomon_)

    #  check bounds consistency
    z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = helpers.get_full_outputs_from_outputs_for_mode(
        outputs_for_mode=outputs_, mode=mode, dc_decomp=dc_decomp, full_inputs=inputs_
    )
    helpers.assert_output_properties_box(
        input_ref_reshaped_,
        output_ref_,
        h_,
        g_,
        input_ref_min_reshaped_,
        input_ref_max_reshaped_,
        u_c_,
        w_u_,
        b_u_,
        l_c_,
        w_l_,
        b_l_,
        decimal=decimal,
    )



from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import (
    Activation,
    Concatenate,
    Flatten,
    Input,
    Lambda,
    Layer,
    Maximum,
    Minimum,
)
from tensorflow.keras.models import Model

from decomon.core import (
    BallDomain,
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    get_mode,
)
from decomon.layers.utils import is_a_merge_layer

try:
    from keras.src.engine.node import Node  # new path starting from keras 2.13
except ImportError:
    from keras.engine.node import Node  # old path until keras 2.12

try:
    from keras.src.backend import (
        observe_object_name,  # new path starting from keras 2.13
    )
except ImportError:
    from keras.backend import observe_object_name  # old path until keras 2.12

try:
    from deel.lip.layers import LipschitzLayer
except ImportError:
    LipschitzLayer = type(None)


class ConvertMethod(Enum):
    CROWN = "crown"
    CROWN_FORWARD_IBP = "crown-forward-ibp"
    CROWN_FORWARD_AFFINE = "crown-forward-affine"
    CROWN_FORWARD_HYBRID = "crown-forward-hybrid"
    FORWARD_IBP = "forward-ibp"
    FORWARD_AFFINE = "forward-affine"
    FORWARD_HYBRID = "forward-hybrid"


def get_ibp_affine_from_method(method: Union[str, ConvertMethod]) -> Tuple[bool, bool]:
    method = ConvertMethod(method)
    if method in [ConvertMethod.FORWARD_IBP, ConvertMethod.CROWN_FORWARD_IBP]:
        return True, False
    if method in [ConvertMethod.FORWARD_AFFINE, ConvertMethod.CROWN_FORWARD_AFFINE]:
        return False, True
    if method in [ConvertMethod.FORWARD_HYBRID, ConvertMethod.CROWN_FORWARD_HYBRID]:
        return True, True
    if method == ConvertMethod.CROWN:
        return True, False
    return True, True


class FeedDirection(Enum):
    FORWARD = "feed_forward"
    BACKWARD = "feed_backward"


def get_direction(method: Union[str, ConvertMethod]) -> FeedDirection:
    if ConvertMethod(method) in [ConvertMethod.FORWARD_IBP, ConvertMethod.FORWARD_AFFINE, ConvertMethod.FORWARD_HYBRID]:
        return FeedDirection.FORWARD
    else:
        return FeedDirection.BACKWARD


def has_merge_layers(model: Model) -> bool:
    return any(is_a_merge_layer(layer) for layer in model.layers)


def get_input_tensors(
    model: Model,
    perturbation_domain: PerturbationDomain,
    ibp: bool = True,
    affine: bool = True,
) -> Tuple[tf.Tensor, List[tf.Tensor]]:
    input_dim = get_input_dim(model)
    input_shape = None
    input_shape_vec = None

    for input_layer in model._input_layers:
        if len(input_layer.input_shape) > 1:
            raise ValueError(f"Expected one input tensor but got {len(input_layer.input_shape)}")
        input_shape_vec_default = input_layer.input_shape[0]
        input_shape_default = tuple(list(input_shape_vec_default)[1:])

        if input_shape_vec is None:
            input_shape_vec = input_shape_vec_default
        if input_shape is None:
            input_shape = input_shape_default
        else:
            if not np.allclose(input_shape, input_shape_default):
                raise ValueError("Expected that every input layers use the same input_tensor")

    mode = get_mode(ibp=ibp, affine=affine)
    dc_decomp = False
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    empty_tensor = inputs_outputs_spec.get_empty_tensor()

    input_shape_x = perturbation_domain.get_x_input_shape(input_dim)
    z_tensor = Input(shape=input_shape_x, dtype=model.layers[0].dtype)
    u_c_tensor, l_c_tensor, W, b, h, g = (
        empty_tensor,
        empty_tensor,
        empty_tensor,
        empty_tensor,
        empty_tensor,
        empty_tensor,
    )

    if isinstance(perturbation_domain, BoxDomain):
        if ibp:
            u_c_tensor = Lambda(lambda z: z[:, 1], dtype=z_tensor.dtype)(z_tensor)
            l_c_tensor = Lambda(lambda z: z[:, 0], dtype=z_tensor.dtype)(z_tensor)

        if affine:
            z_value = K.cast(0.0, z_tensor.dtype)
            o_value = K.cast(1.0, z_tensor.dtype)
            W = Lambda(lambda z: tf.linalg.diag(z_value * z[:, 0] + o_value), dtype=z_tensor.dtype)(z_tensor)
            b = Lambda(lambda z: z_value * z[:, 1], dtype=z_tensor.dtype)(z_tensor)

    elif isinstance(perturbation_domain, BallDomain):
        if perturbation_domain.p == np.inf:
            radius = perturbation_domain.eps
            if ibp:
                u_c_tensor = Lambda(
                    lambda var: var + K.cast(radius, dtype=model.layers[0].dtype), dtype=model.layers[0].dtype
                )(z_tensor)
                l_c_tensor = Lambda(
                    lambda var: var - K.cast(radius, dtype=model.layers[0].dtype), dtype=model.layers[0].dtype
                )(z_tensor)
            if affine:
                z_value = K.cast(0.0, model.layers[0].dtype)
                o_value = K.cast(1.0, model.layers[0].dtype)

                W = tf.linalg.diag(z_value * u_c_tensor + o_value)
                b = z_value * u_c_tensor

        else:
            z_value = K.cast(0.0, model.layers[0].dtype)
            o_value = K.cast(1.0, model.layers[0].dtype)

            def get_bounds(z: tf.Tensor) -> List[tf.Tensor]:
                outputs = []
                W = tf.linalg.diag(z_value * z + o_value)
                b = z_value * z
                if affine:
                    outputs += [W, b]
                if ibp:
                    u_c_out = perturbation_domain.get_upper(z, W, b)
                    l_c_out = perturbation_domain.get_lower(z, W, b)
                    outputs += [u_c_out, l_c_out]
                return outputs

            outputs = get_bounds(z_tensor)
            if ibp:
                u_c_tensor, l_c_tensor = outputs[-2:]
            if affine:
                W, b = outputs[:2]
    else:
        raise NotImplementedError(f"Not implemented for perturbation domain type {type(perturbation_domain)}")

    input_tensors = inputs_outputs_spec.extract_inputsformode_from_fullinputs(
        [z_tensor, u_c_tensor, W, b, l_c_tensor, W, b, h, g]
    )

    return z_tensor, input_tensors


def get_input_tensors_keras_only(
    model: Model,
    input_shape: Tuple[int, ...],
) -> List[tf.Tensor]:
    input_tensors = []
    for i in range(len(model._input_layers)):
        input_tensors.append(Input(input_shape[1:], dtype=model.layers[0].dtype))
    return input_tensors


def get_input_dim(layer: Layer) -> int:
    if isinstance(layer.input_shape, list):
        return np.prod(layer.input_shape[0][1:])
    else:
        return np.prod(layer.input_shape[1:])


def prepare_inputs_for_layer(
    inputs: Union[Tuple[tf.Tensor, ...], List[tf.Tensor], tf.Tensor]
) -> Union[Tuple[tf.Tensor, ...], List[tf.Tensor], tf.Tensor]:
    """Prepare inputs for keras/decomon layers.

    Some Keras layers do not like list of tensors even with one single tensor.
    So we keep only the tensor in this case.

    """
    if isinstance(inputs, list) or isinstance(inputs, tuple):
        if len(inputs) == 1:
            return inputs[0]
    return inputs


def wrap_outputs_from_layer_in_list(
    outputs: Union[Tuple[tf.Tensor, ...], List[tf.Tensor], tf.Tensor]
) -> List[tf.Tensor]:
    if not isinstance(outputs, list):
        if isinstance(outputs, tuple):
            return list(outputs)
        else:
            return [outputs]
    else:
        return outputs


def _share_no_attributes(new_layer: Layer, old_layer: Layer) -> None:
    return


def split_activation(
    layer: Layer, share_some_attributes: Optional[Callable[[Layer, Layer], None]] = None
) -> List[Layer]:
    # init
    if share_some_attributes is None:
        share_some_attributes = _share_no_attributes
    # get activation
    config = layer.get_config()
    activation = config.pop("activation", None)
    # split if needed
    if activation is None or activation == "linear" or isinstance(layer, Activation):
        return [layer]
    else:
        # layer without activation
        config["name"] = f"{layer.name}_wo_activation"
        layer_wo_activation = layer.__class__.from_config(config)
        share_some_attributes(layer_wo_activation, layer)  # share (deel-lip) attributes
        # build the layer
        if not hasattr(layer, "input_shape"):
            raise RuntimeError("The layer should properly initialized so that layer.input_shape is defined.")
        inputs = Input(type_spec=layer.input.type_spec)
        outputs = layer_wo_activation(inputs)
        # use same weights
        actual_activation = layer.activation  # store activation
        layer.activation = None  # remove activation in case of a layer having weights (would be in layer.get_weights())
        layer_wo_activation.set_weights(layer.get_weights())
        layer.activation = actual_activation  # put back the actual activation
        # even share the object themselves if possible
        # (dense and conv2d have kernel and bias weight attributes)
        if hasattr(layer_wo_activation, "kernel"):
            layer_wo_activation.kernel = layer.kernel
        if hasattr(layer_wo_activation, "bias"):
            layer_wo_activation.bias = layer.bias
        # activation layer
        if isinstance(activation, dict):
            if isinstance(layer.activation, Layer):  # can be an Activation, a PReLU, or a deel.lip.activations layer
                activation_layer = layer.activation
                # update the name to starts with main layer name
                activation_layer_name = f"{layer.name}_activation_{layer.activation.name}"
                observe_object_name(activation_layer_name)
                activation_layer._name = activation_layer_name
            else:
                raise RuntimeError("Cannot construct activation layer from layer.activation!")
        else:
            activation_layer = Activation(activation=activation, dtype=layer.dtype, name=f"{layer.name}_activation")
        # build activation layer
        activation_layer(outputs)
        return [layer_wo_activation, activation_layer]


def convert_deellip_to_keras(layer: Layer) -> Layer:
    # init deel-lip attributes (keep exisiting ones)
    share_deellip_attributes(layer, layer)
    # update is_lipschitz
    if isinstance(layer, LipschitzLayer) or hasattr(layer, "vanilla_export"):
        layer.is_lipschitz = True
    if hasattr(layer, "vanilla_export"):
        if not hasattr(layer, "input_shape"):
            raise RuntimeError("The layer should properly initialized so that layer.input_shape is defined.")
        type_spec = layer.input.type_spec
        new_layer = layer.vanilla_export()
        # build layer
        inputs = Input(type_spec=type_spec)
        new_layer(inputs)
        # share deel-lip attributes of original layer
        share_deellip_attributes(new_layer, layer)
        layer = new_layer

    return layer


def share_deellip_attributes(new_layer: Layer, old_layer: Layer = None) -> None:
    new_layer.is_lipschitz = getattr(old_layer, "is_lipschitz", False)
    new_layer.deellip_classname = getattr(old_layer, "deellip_classname", new_layer.__class__.__name__)
    new_layer.k_coef_lip = getattr(old_layer, "k_coef_lip", -1.0)


def preprocess_layer(layer: Layer) -> List[Layer]:
    layer = convert_deellip_to_keras(layer)
    layers = split_activation(layer, share_some_attributes=share_deellip_attributes)
    # convert activation layers (if were embedded deel-lip layers)
    return [convert_deellip_to_keras(l) for l in layers]


def get_depth_dict(model: Model) -> Dict[int, List[Node]]:
    depth_keys = list(model._nodes_by_depth.keys())
    depth_keys.sort(reverse=True)

    nodes_list = []

    dico_depth: Dict[int, int] = {}
    dico_nodes: Dict[int, List[Node]] = {}

    def fill_dico(node: Node, dico_depth: Optional[Dict[int, int]] = None) -> Dict[int, int]:
        if dico_depth is None:
            dico_depth = {}

        parents = node.parent_nodes
        if len(parents):
            for parent in parents:
                if id(parent) not in dico_depth:
                    dico_depth = fill_dico(parent, dico_depth)

                depth = np.min([dico_depth[id(parent)] - 1 for parent in parents])
                if id(node) in dico_depth:
                    dico_depth[id(node)] = max(depth, dico_depth[id(node)])
                else:
                    dico_depth[id(node)] = depth
        else:
            dico_depth[id(node)] = max(depth_keys)

        return dico_depth

    for depth in depth_keys:
        # check for nodes that do not have parents and set depth to maximum depth value
        nodes = model._nodes_by_depth[depth]
        nodes_list += nodes
        for node in nodes:
            dico_depth = fill_dico(node, dico_depth)

    for node in nodes_list:
        depth = dico_depth[id(node)]
        if depth in dico_nodes:
            dico_nodes[depth].append(node)
        else:
            dico_nodes[depth] = [node]

    return dico_nodes


def get_inner_layers(model: Model) -> int:
    count = 0
    for layer in model.layers:
        if isinstance(layer, Model):
            count += get_inner_layers(layer)
        else:
            count += 1
    return count


class Convert2Mode(Layer):
    def __init__(
        self,
        mode_from: Union[str, ForwardMode],
        mode_to: Union[str, ForwardMode],
        perturbation_domain: PerturbationDomain,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.mode_from = ForwardMode(mode_from)
        self.mode_to = ForwardMode(mode_to)
        self.perturbation_domain = perturbation_domain

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        mode_from = self.mode_from
        mode_to = self.mode_to
        perturbation_domain = self.perturbation_domain

        dc_decomp = False
        inputs_outputs_spec_from = InputsOutputsSpec(
            dc_decomp=dc_decomp, mode=mode_from, perturbation_domain=perturbation_domain
        )
        inputs_outputs_spec_to = InputsOutputsSpec(
            dc_decomp=dc_decomp, mode=mode_to, perturbation_domain=perturbation_domain
        )

        compute_ibp_from_affine = mode_from == ForwardMode.AFFINE and mode_to != ForwardMode.AFFINE
        tight = mode_from == ForwardMode.HYBRID and mode_to != ForwardMode.AFFINE
        compute_dummy_affine = mode_from == ForwardMode.IBP and mode_to != ForwardMode.IBP
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec_from.get_fullinputs_from_inputsformode(
            inputs, compute_ibp_from_affine=compute_ibp_from_affine, tight=tight
        )
        dtype = x.dtype

        if compute_dummy_affine:
            x = K.concatenate([K.expand_dims(l_c, 1), K.expand_dims(u_c, 1)], 1)
            z_value = K.cast(0.0, dtype=dtype)
            w = tf.linalg.diag(z_value * l_c)
            w_u = w
            b_u = u_c
            w_l = w
            b_l = l_c

        return inputs_outputs_spec_to.extract_outputsformode_from_fulloutputs([x, u_c, w_u, b_u, l_c, w_l, b_l, h, g])

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {"mode_from": self.mode_from, "mode_to": self.mode_to, "perturbation_domain": self.perturbation_domain}
        )
        return config



import logging
from typing import Any, Dict, Optional, Union

from tensorflow.keras.layers import Layer

from decomon.backward_layers.core import BackwardLayer
from decomon.core import ForwardMode, PerturbationDomain, get_affine, get_ibp
from decomon.layers.convert import to_decomon
from decomon.layers.decomon_layers import DecomonLayer

logger = logging.getLogger(__name__)

try:
    from deel.lip.activations import GroupSort, GroupSort2
except ImportError:
    logger.warning(
        "Could not import GroupSort or GroupSort2 from deel.lip.activations. "
        "Please install deel-lip for being compatible with 1 Lipschitz network (see https://github.com/deel-ai/deel-lip)"
    )


class BackwardGroupSort2(BackwardLayer):
    """Backward LiRPA of GroupSort2"""

    def __init__(
        self,
        layer: Layer,
        input_dim: int = -1,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )

        if not isinstance(layer, DecomonLayer):
            self.layer = to_decomon(
                layer=layer,
                input_dim=input_dim,
                dc_decomp=False,
                perturbation_domain=self.perturbation_domain,
                finetune=False,
                ibp=get_ibp(self.mode),
                affine=get_affine(self.mode),
                shared=True,
                fast=False,
            )

        self.frozen_weights = False



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Type, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    get_affine,
    get_ibp,
)
from decomon.keras_utils import get_weight_index_from_name


class DecomonLayer(ABC, Layer):
    """Abstract class that contains the common information of every implemented layers for Forward LiRPA"""

    _trainable_weights: List[tf.Variable]

    @property
    @abstractmethod
    def original_keras_layer_class(self) -> Type[Layer]:
        """The keras layer class from which this class is the decomon equivalent."""
        pass

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        """
        Args:
            perturbation_domain: type of convex input domain (None or dict)
            dc_decomp: boolean that indicates whether we return a
            mode: type of Forward propagation (ibp, affine, or hybrid)
            **kwargs: extra parameters
        difference of convex decomposition of our layer
        """
        kwargs.pop("slope", None)  # remove it if not used by the decomon layer
        super().__init__(**kwargs)

        if perturbation_domain is None:
            perturbation_domain = BoxDomain()
        self.inputs_outputs_spec = InputsOutputsSpec(
            dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain
        )
        self.nb_tensors = self.inputs_outputs_spec.nb_tensors
        self.dc_decomp = dc_decomp
        self.perturbation_domain = perturbation_domain
        self.mode = ForwardMode(mode)
        self.finetune = finetune  # extra optimization with hyperparameters
        self.frozen_weights = False
        self.frozen_alpha = False
        self.shared = shared
        self.fast = fast
        self.has_backward_bounds = False  # optimizing Forward LiRPA for adversarial perturbation

    @property
    def ibp(self) -> bool:
        return get_ibp(self.mode)

    @property
    def affine(self) -> bool:
        return get_affine(self.mode)

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "dc_decomp": self.dc_decomp,
                "shared": self.shared,
                "fast": self.fast,
                "finetune": self.finetune,
                "mode": self.mode,
                "perturbation_domain": self.perturbation_domain,
            }
        )
        return config

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        """
        Args:
            input_shape

        Returns:

        """
        # generic case: call build from underlying keras layer with the proper intput_shape
        y_input_shape = input_shape[-1]
        self.original_keras_layer_class.build(self, y_input_shape)

    @abstractmethod
    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        """
        Args:
            inputs

        Returns:

        """

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> List[tf.TensorShape]:
        """
        Args:
            input_shape

        Returns:

        """
        return self.original_keras_layer_class.compute_output_shape(self, input_shape)

    def reset_layer(self, layer: Layer) -> None:
        """Reset the weights by using the weights of another (a priori non-decomon) layer.

        It set the weights whose names are listed by `keras_weights_names`.

        Args:
            layer

        Returns:

        """
        weight_names = self.keras_weights_names
        if len(weight_names) > 0:
            reset_layer(decomon_layer=self, keras_layer=layer, weight_names=weight_names)

    @property
    def keras_weights_names(self) -> List[str]:
        """Weights names of the corresponding Keras layer.

        Will be used to decide which weight to take from the keras layer in `reset_layer()`

        """
        return []

    def join(self, bounds: List[tf.Tensor]) -> List[tf.Tensor]:
        """
        Args:
            bounds

        Returns:

        """
        raise NotImplementedError()

    def freeze_weights(self) -> None:
        pass

    def unfreeze_weights(self) -> None:
        pass

    def freeze_alpha(self) -> None:
        pass

    def unfreeze_alpha(self) -> None:
        pass

    def reset_finetuning(self) -> None:
        pass

    def share_weights(self, layer: Layer) -> None:
        pass

    def split_kwargs(self, **kwargs: Any) -> None:
        # necessary for InputLayer
        pass

    def set_back_bounds(self, has_backward_bounds: bool) -> None:
        pass


def reset_layer(decomon_layer: DecomonLayer, keras_layer: Layer, weight_names: List[str]) -> None:
    """Reset the weights of a decomon layer by using the weights of another (a priori non-decomon) layer.

    Args:
        decomon_layer: the decomon layer whose weights will be updated
        keras_layer: the layer used to update the weights
        weight_names: the names of the weights to update

    Returns:

    """
    if not keras_layer.built:
        raise ValueError(f"the layer {keras_layer.name} has not been built yet")
    if not decomon_layer.built:
        raise ValueError(f"the layer {decomon_layer.name} has not been built yet")
    else:
        decomon_params = decomon_layer.get_weights()
        original_params = keras_layer.get_weights()
        for weight_name in weight_names:
            decomon_params[get_weight_index_from_name(decomon_layer, weight_name)] = original_params[
                get_weight_index_from_name(keras_layer, weight_name)
            ]
        decomon_layer.set_weights(decomon_params)



from abc import ABC, abstractmethod
from itertools import chain
from typing import Any, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer, Wrapper

from decomon.backward_layers.utils import (
    backward_add,
    backward_maximum,
    backward_minimum,
    backward_multiply,
    backward_subtract,
    get_identity_lirpa,
)
from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    get_affine,
    get_ibp,
)
from decomon.layers.core import DecomonLayer
from decomon.layers.decomon_merge_layers import (
    DecomonAdd,
    DecomonConcatenate,
    DecomonDot,
    DecomonMaximum,
    DecomonMinimum,
    DecomonMultiply,
    DecomonSubtract,
)
from decomon.layers.utils import broadcast, multiply, permute_dimensions, split


class BackwardMerge(ABC, Wrapper):
    layer: Layer
    _trainable_weights: List[tf.Variable]

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        kwargs.pop("slope", None)
        kwargs.pop("finetune", None)
        super().__init__(layer, **kwargs)
        self.rec = rec
        if isinstance(self.layer, DecomonLayer):
            self.mode = self.layer.mode
            self.perturbation_domain = self.layer.perturbation_domain
            self.dc_decomp = self.layer.dc_decomp
        else:
            self.mode = ForwardMode(mode)
            if perturbation_domain is None:
                self.perturbation_domain = BoxDomain()
            else:
                self.perturbation_domain = perturbation_domain
            self.dc_decomp = dc_decomp
        self.inputs_outputs_spec = InputsOutputsSpec(
            dc_decomp=self.dc_decomp, mode=self.mode, perturbation_domain=self.perturbation_domain
        )

    @property
    def ibp(self) -> bool:
        return get_ibp(self.mode)

    @property
    def affine(self) -> bool:
        return get_affine(self.mode)

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "rec": self.rec,
                "mode": self.mode,
                "perturbation_domain": self.perturbation_domain,
                "dc_decomp": self.dc_decomp,
            }
        )
        return config

    @abstractmethod
    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        """
        Args:
            inputs

        Returns:

        """
        pass

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        """
        Args:
            input_shape

        Returns:

        """
        # generic case: nothing to do before call
        pass

    def freeze_weights(self) -> None:
        pass

    def unfreeze_weights(self) -> None:
        pass

    def freeze_alpha(self) -> None:
        pass

    def unfreeze_alpha(self) -> None:
        pass

    def reset_finetuning(self) -> None:
        pass


class BackwardAdd(BackwardMerge):
    """Backward  LiRPA of Add"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        self.op = DecomonAdd(
            mode=self.mode, perturbation_domain=self.perturbation_domain, dc_decomp=self.dc_decomp
        ).call

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        #  check number of inputs
        if len(inputs_list) == 1:  # nothing to merge
            return [[w_u_out, b_u_out, w_l_out, b_l_out]]
        elif len(inputs_list) == 2:
            bounds_0, bounds_1 = backward_add(
                inputs_list[0],
                inputs_list[1],
                w_u_out,
                b_u_out,
                w_l_out,
                b_l_out,
                perturbation_domain=self.perturbation_domain,
                mode=self.mode,
                dc_decomp=self.dc_decomp,
            )
            return [bounds_0, bounds_1]

        else:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")


class BackwardAverage(BackwardMerge):
    """Backward  LiRPA of Average"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        self.op = DecomonAdd(mode=self.mode, perturbation_domain=self.perturbation_domain, dc_decomp=False).call

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)
        #  check number of inputs
        if n_elem == 1:  # nothing to merge
            return [[w_u_out, b_u_out, w_l_out, b_l_out]]
        else:
            bounds: List[List[tf.Tensor]] = []
            input_bounds: List[List[tf.Tensor]] = []

            for j in range(n_elem - 1, 0, -1):
                inputs_1 = inputs_list[j]
                if j == 1:
                    inputs_0 = inputs_list[0]
                else:
                    inputs_0 = self.op(list(chain(*inputs_list[: (j - 1)])))  # merge (j-1) first inputs
                bounds_0, bounds_1 = backward_add(
                    inputs_0,
                    inputs_1,
                    w_u_out,
                    b_u_out,
                    w_l_out,
                    b_l_out,
                    perturbation_domain=self.perturbation_domain,
                    mode=self.mode,
                    dc_decomp=self.dc_decomp,
                )
                input_bounds.append(bounds_1)
                bounds.append(bounds_0)
                if j == 1:
                    input_bounds.append(bounds_0)
                # update bounds to use for next iteration
                w_u_out, b_u_out, w_l_out, b_l_out = bounds_0

            input_bounds = input_bounds[::-1]
            return [[1.0 / n_elem * elem_i for elem_i in elem] for elem in input_bounds]


class BackwardSubtract(BackwardMerge):
    """Backward  LiRPA of Subtract"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonSubtract):
            raise KeyError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)

        if n_elem != 2:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")

        return backward_subtract(
            inputs_list[0],
            inputs_list[1],
            w_u_out,
            b_u_out,
            w_l_out,
            b_l_out,
            perturbation_domain=self.layer.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
        )


class BackwardMaximum(BackwardMerge):
    """Backward  LiRPA of Maximum"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonMaximum):
            raise KeyError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)

        if n_elem != 2:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")

        return backward_maximum(
            inputs_list[0],
            inputs_list[1],
            w_u_out,
            b_u_out,
            w_l_out,
            b_l_out,
            perturbation_domain=self.layer.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
        )


class BackwardMinimum(BackwardMerge):
    """Backward  LiRPA of Minimum"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonMinimum):
            raise KeyError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)

        if n_elem != 2:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")

        return backward_minimum(
            inputs_list[0],
            inputs_list[1],
            w_u_out,
            b_u_out,
            w_l_out,
            b_l_out,
            perturbation_domain=self.layer.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
        )


class BackwardConcatenate(BackwardMerge):
    """Backward  LiRPA of Concatenate"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonConcatenate):
            raise KeyError()

        self.axis = self.layer.axis

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)
        n_list = [self.inputs_outputs_spec.get_input_shape(subinputs)[self.axis] for subinputs in inputs_list]
        axis_w = self.axis
        if axis_w != -1:
            axis_w += 1
        w_u_out_list = tf.split(w_u_out, n_list, axis_w)
        w_l_out_list = tf.split(w_l_out, n_list, axis_w)
        b_u_out_list = tf.split(b_u_out, n_list, self.axis)
        b_l_out_list = tf.split(b_l_out, n_list, self.axis)

        bounds = [[w_u_out_list[i], b_u_out_list[i], w_l_out_list[i], b_l_out_list[i]] for i in range(n_elem)]

        return bounds


class BackwardMultiply(BackwardMerge):
    """Backward  LiRPA of Multiply"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonMultiply):
            raise KeyError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)

        if n_elem != 2:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")

        return backward_multiply(
            inputs_list[0],
            inputs_list[1],
            w_u_out,
            b_u_out,
            w_l_out,
            b_l_out,
            perturbation_domain=self.layer.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
        )


class BackwardDot(BackwardMerge):
    """Backward  LiRPA of Dot"""

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        if not isinstance(layer, DecomonDot):
            raise KeyError()

        self.axes = [i for i in self.layer.axes]
        self.op = BackwardAdd(self.layer)

        raise NotImplementedError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[List[tf.Tensor]]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)

        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        n_elem = len(inputs_list)

        if n_elem != 2:
            raise NotImplementedError("This layer is intended to merge only 2 layers.")

        # permute dimensions and reshape
        inputs_0 = inputs_list[0]
        inputs_1 = inputs_list[1]

        n_0 = len(inputs_0[0].shape) - 2
        n_1 = len(inputs_1[0].shape) - 2

        input_0_permuted = permute_dimensions(inputs_0, self.axes[0], mode=self.mode, dc_decomp=self.dc_decomp)
        input_1_permuted = permute_dimensions(inputs_1, self.axes[1], mode=self.mode, dc_decomp=self.dc_decomp)

        inputs_0_broadcasted = broadcast(input_0_permuted, n_1, -1, mode=self.mode, dc_decomp=self.dc_decomp)
        inputs_1_broadcasted = broadcast(input_1_permuted, n_0, 2, mode=self.mode, dc_decomp=self.dc_decomp)

        inputs_multiplied = multiply(
            inputs_0_broadcasted,
            inputs_1_broadcasted,
            dc_decomp=self.layer.dc_decomp,
            perturbation_domain=self.perturbation_domain,
            mode=self.mode,
        )

        inputs_add = []
        for elem in split(inputs_multiplied, axis=self.axes[0], mode=self.mode):
            inputs_add += elem

        bounds = self.op.call(inputs_add)
        n = len(inputs_add)
        bounds = [[1.0 / n * elem for elem in bounds_i] for bounds_i in bounds]

        # concatenate
        shape = [-1, 1] + list(inputs_multiplied[0].shape[1:]) + list(w_u_out.shape[3:])

        bounds_reshaped = [[K.reshape(elem[0], shape), elem[1], K.reshape(elem[2], shape), elem[3]] for elem in bounds]
        w_u = K.concatenate([elem[0] for elem in bounds_reshaped], 1)
        b_u = sum([elem[1] for elem in bounds_reshaped], 1)
        w_l = K.concatenate([elem[2] for elem in bounds_reshaped], 1)
        b_l = sum([elem[3] for elem in bounds_reshaped])

        bounds_m_0, bounds_m_1 = backward_multiply(
            inputs_0_broadcasted,
            inputs_1_broadcasted,
            w_u,
            b_u,
            w_l,
            b_l,
            perturbation_domain=self.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
        )

        shape_0 = [-1, 1] + list(input_0_permuted[0].shape[1:]) + list(w_u.shape[3:])
        shape_1 = [-1, 1] + list(input_1_permuted[0].shape[1:]) + list(w_u.shape[3:])

        bounds_m_0 = [
            K.reshape(bounds_m_0[0], shape_0),
            bounds_m_0[1],
            K.reshape(bounds_m_0[2], shape_0),
            bounds_m_0[3],
        ]
        bounds_m_1 = [
            K.reshape(bounds_m_1[0], shape_1),
            bounds_m_1[1],
            K.reshape(bounds_m_1[2], shape_1),
            bounds_m_1[3],
        ]

        axes = [i for i in self.axes]
        if axes[0] == -1:
            axes[0] = len(inputs_0[0].shape)
        if axes[1] == -1:
            axes[1] = len(inputs_1[0].shape)

        index_0 = np.arange(len(shape_0))
        index_0[2] = axes[0] + 1
        index_0[axes[0] + 1] = 2

        index_1 = np.arange(len(shape_1))
        index_1[2] = axes[1] + 1
        index_1[axes[1] + 1] = 2

        bounds_m_0 = [
            K.permute_dimensions(bounds_m_0[0], index_0),
            bounds_m_0[1],
            K.permute_dimensions(bounds_m_0[2], index_0),
            bounds_m_0[3],
        ]
        bounds_m_1 = [
            K.permute_dimensions(bounds_m_1[0], index_1),
            bounds_m_1[1],
            K.permute_dimensions(bounds_m_1[2], index_1),
            bounds_m_1[3],
        ]

        return [bounds_m_0, bounds_m_1]



import logging
from typing import Any, Dict, List, Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

from decomon.core import ForwardMode, PerturbationDomain
from decomon.layers.core import DecomonLayer
from decomon.layers.decomon_merge_layers import DecomonConcatenate
from decomon.layers.decomon_reshape import DecomonReshape
from decomon.layers.utils import ClipAlpha, expand_dims, max_, min_, sort

logger = logging.getLogger(__name__)

try:
    from deel.lip.activations import GroupSort, GroupSort2
except ImportError:
    logger.warning(
        "Could not import GroupSort or GroupSort2 from deel.lip.activations. "
        "Please install deel-lip for being compatible with 1 Lipschitz network (see https://github.com/deel-ai/deel-lip)"
    )
else:

    class DecomonGroupSort(DecomonLayer):
        original_keras_layer_class = GroupSort

        def __init__(
            self,
            n: Optional[int] = None,
            data_format: str = "channels_last",
            k_coef_lip: float = 1.0,
            perturbation_domain: Optional[PerturbationDomain] = None,
            dc_decomp: bool = False,
            mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
            finetune: bool = False,
            shared: bool = False,
            fast: bool = True,
            **kwargs: Any,
        ):
            super().__init__(
                perturbation_domain=perturbation_domain,
                dc_decomp=dc_decomp,
                mode=mode,
                finetune=finetune,
                shared=shared,
                fast=fast,
                **kwargs,
            )
            self.data_format = data_format
            if data_format == "channels_last":
                self.channel_axis = -1
            elif data_format == "channels_first":
                raise RuntimeError("channels_first not implemented for GroupSort activation")
            else:
                raise RuntimeError("data format not understood")
            self.n = n
            self.concat = DecomonConcatenate(
                mode=self.mode, perturbation_domain=self.perturbation_domain, dc_decomp=self.dc_decomp
            ).call

        def get_config(self) -> Dict[str, Any]:
            config = super().get_config()
            config.update(
                {
                    "data_format": self.data_format,
                    "mode": self.mode,
                    "n": self.n,
                }
            )
            return config

        def build(self, input_shape: List[tf.TensorShape]) -> None:
            if (self.n is None) or (self.n > input_shape[-1][self.channel_axis]):
                self.n = input_shape[-1][self.channel_axis]
                if self.n is None:  # for mypy
                    raise RuntimeError("self.n cannot be None at this point.")
            self.reshape = DecomonReshape(
                (-1, self.n), mode=self.mode, perturbation_domain=self.perturbation_domain, dc_decomp=self.dc_decomp
            ).call

        def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
            shape_in = tuple(inputs[-1].shape[1:])
            inputs_reshaped = self.reshape(inputs)
            if self.n == 2:
                output_max = expand_dims(
                    max_(
                        inputs_reshaped,
                        dc_decomp=self.dc_decomp,
                        perturbation_domain=self.perturbation_domain,
                        mode=self.mode,
                        axis=-1,
                    ),
                    dc_decomp=self.dc_decomp,
                    mode=self.mode,
                    axis=-1,
                    perturbation_domain=self.perturbation_domain,
                )
                output_min = expand_dims(
                    min_(
                        inputs_reshaped,
                        dc_decomp=self.dc_decomp,
                        perturbation_domain=self.perturbation_domain,
                        mode=self.mode,
                        axis=-1,
                    ),
                    dc_decomp=self.dc_decomp,
                    mode=self.mode,
                    axis=-1,
                    perturbation_domain=self.perturbation_domain,
                )
                outputs = self.concat(output_min + output_max)

            else:
                outputs = sort(
                    inputs_reshaped,
                    axis=-1,
                    dc_decomp=self.dc_decomp,
                    perturbation_domain=self.perturbation_domain,
                    mode=self.mode,
                )

            return DecomonReshape(
                shape_in, mode=self.mode, perturbation_domain=self.perturbation_domain, dc_decomp=self.dc_decomp
            ).call(outputs)

        def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> List[tf.TensorShape]:
            return input_shape

    class DecomonGroupSort2(DecomonLayer):
        original_keras_layer_class = GroupSort2

        def __init__(
            self,
            n: int = 2,
            data_format: str = "channels_last",
            k_coef_lip: float = 1.0,
            perturbation_domain: Optional[PerturbationDomain] = None,
            dc_decomp: bool = False,
            mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
            finetune: bool = False,
            shared: bool = False,
            fast: bool = True,
            **kwargs: Any,
        ):
            super().__init__(
                perturbation_domain=perturbation_domain,
                dc_decomp=dc_decomp,
                mode=mode,
                finetune=finetune,
                shared=shared,
                fast=fast,
                **kwargs,
            )
            self.data_format = data_format

            if self.data_format == "channels_last":
                self.axis = -1
            else:
                self.axis = 1

            if self.dc_decomp:
                raise NotImplementedError()

            self.op_concat = DecomonConcatenate(self.axis, mode=self.mode, perturbation_domain=self.perturbation_domain)

        def get_config(self) -> Dict[str, Any]:
            config = super().get_config()
            config.update(
                {
                    "data_format": self.data_format,
                    "mode": self.mode,
                }
            )
            return config

        def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> List[tf.TensorShape]:
            return input_shape

        def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
            inputs_reshaped = self.op_reshape_in(inputs)
            inputs_max = expand_dims(
                max_(
                    inputs_reshaped,
                    mode=self.mode,
                    perturbation_domain=self.perturbation_domain,
                    axis=self.axis,
                    finetune=self.finetune,
                    finetune_params=self.params_max,
                ),
                mode=self.mode,
                axis=self.axis,
                perturbation_domain=self.perturbation_domain,
            )
            inputs_min = expand_dims(
                min_(
                    inputs_reshaped,
                    mode=self.mode,
                    perturbation_domain=self.perturbation_domain,
                    axis=self.axis,
                    finetune=self.finetune,
                    finetune_params=self.params_min,
                ),
                mode=self.mode,
                axis=self.axis,
                perturbation_domain=self.perturbation_domain,
            )
            output = self.op_concat(inputs_min + inputs_max)
            return self.op_reshape_out(output)

        def build(self, input_shape: List[tf.TensorShape]) -> None:
            input_shape = input_shape[-1]

            if self.data_format == "channels_last":
                if input_shape[-1] % 2 != 0:
                    raise ValueError()
                target_shape = list(input_shape[1:-2]) + [int(input_shape[-1] / 2), 2]
            else:
                if input_shape[1] % 2 != 0:
                    raise ValueError()
                target_shape = [2, int(input_shape[1] / 2)] + list(input_shape[2:])

            self.params_max = []
            self.params_min = []

            if self.finetune and self.mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
                self.beta_max = self.add_weight(
                    shape=target_shape, initializer="ones", name="beta_max", regularizer=None, constraint=ClipAlpha()
                )
                self.beta_min = self.add_weight(
                    shape=target_shape, initializer="ones", name="beta_max", regularizer=None, constraint=ClipAlpha()
                )
                self.params_max = [self.beta_max]
                self.params_min = [self.beta_min]

            self.op_reshape_in = DecomonReshape(tuple(target_shape), mode=self.mode)
            self.op_reshape_out = DecomonReshape(tuple(input_shape[1:]), mode=self.mode)

        def reset_layer(self, layer: Layer) -> None:
            pass



import pytest
from tensorflow.keras.layers import Add, Conv2D, Dense, Input, Layer, Reshape

from decomon.layers.convert import to_decomon
from decomon.layers.utils import is_a_merge_layer


class MyLayer(Layer):
    """Mock layer unknown from decomon."""

    ...


class MyMerge(Layer):
    """Mock merge layer unknown from decomon."""

    def _merge_function(self, inputs):
        return inputs


def test_is_merge_layer():
    layer = MyMerge()
    assert is_a_merge_layer(layer)
    layer = MyLayer()
    assert not is_a_merge_layer(layer)


def test_to_decomon_merge_not_built_ko():
    layer = MyMerge()
    with pytest.raises(ValueError):
        to_decomon(layer, input_dim=1)


def test_to_decomon_not_built_ko():
    layer = MyLayer()
    with pytest.raises(ValueError):
        to_decomon(layer, input_dim=1)


def test_to_decomon_merge_not_implemented_ko():
    layer = MyMerge()
    layer.built = True
    with pytest.raises(NotImplementedError):
        to_decomon(layer, input_dim=1)


def test_to_decomon_not_implemented_ko():
    layer = MyLayer()
    layer.built = True
    with pytest.raises(NotImplementedError):
        to_decomon(layer, input_dim=1)


@pytest.mark.parametrize(
    "layer_class, layer_kwargs, input_shape_wo_batchsize, nb_inputs",
    [
        (Dense, {"units": 3, "use_bias": True}, (1,), 1),
        (Conv2D, {"filters": 2, "kernel_size": (3, 3), "use_bias": True}, (64, 64, 3), 1),
        (Dense, {"units": 3, "use_bias": False}, (1,), 1),
        (Conv2D, {"filters": 2, "kernel_size": (3, 3), "use_bias": False}, (64, 64, 3), 1),
        (Reshape, {"target_shape": (72,)}, (6, 6, 2), 1),
        (Add, {}, (1,), 2),
    ],
)
def test_to_decomon_ok(layer_class, layer_kwargs, input_shape_wo_batchsize, nb_inputs):
    layer = layer_class(**layer_kwargs)
    # init input_shape and weights
    input_shape = (None,) + input_shape_wo_batchsize
    # input_tensors + build layer
    if nb_inputs == 1:
        input_tensor = Input(input_shape)
        layer(input_tensor)
    else:
        input_tensors = [Input(input_shape) for _ in range(nb_inputs)]
        layer(input_tensors)
    decomon_layer = to_decomon(layer, input_dim=1)
    # check trainable weights
    for i in range(len(layer._trainable_weights)):
        assert layer._trainable_weights[i] is decomon_layer._trainable_weights[i]



from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Lambda, Layer

from decomon.core import (
    BallDomain,
    BoxDomain,
    ForwardMode,
    PerturbationDomain,
    get_mode,
)
from decomon.layers.activations import softmax as softmax_
from decomon.layers.core import DecomonLayer
from decomon.models.models import DecomonModel
from decomon.models.utils import Convert2Mode


def get_model(model: DecomonModel) -> DecomonModel:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    inputs = model.inputs
    outputs = model.outputs
    new_output: tf.Tensor

    if mode == ForwardMode.IBP:

        def func(outputs: List[tf.Tensor]) -> tf.Tensor:
            u_c, l_c = outputs
            return K.concatenate([K.expand_dims(u_c, -1), K.expand_dims(l_c, -1)], -1)

        new_output = Lambda(func)(outputs)

    elif mode == ForwardMode.AFFINE:

        def func(outputs: List[tf.Tensor]) -> tf.Tensor:
            x_0, w_u, b_u, w_l, b_l = outputs
            if len(x_0.shape) == 2:
                x_0_reshaped = x_0[:, :, None]
            else:
                x_0_reshaped = K.permute_dimensions(x_0, (0, 2, 1))
            x_fake = K.sum(x_0_reshaped, 1)[:, None]
            x_0_reshaped = K.concatenate([x_0_reshaped, x_fake], 1)  # (None, n_in+1, n_comp)

            w_b_u = K.concatenate([w_u, b_u[:, None]], 1)  # (None, n_in+1, n_out)
            w_b_l = K.concatenate([w_l, b_l[:, None]], 1)

            w_b = K.concatenate([w_b_u, w_b_l], -1)  # (None, n_in+1, 2*n_out)
            return K.concatenate([x_0_reshaped, w_b], -1)  # (None, n_in+1, n_comp+2*n_out)

        new_output = Lambda(func)(outputs)

    elif mode == ForwardMode.HYBRID:

        def func(outputs: List[tf.Tensor]) -> tf.Tensor:
            x_0, u_c, w_u, b_u, l_c, w_l, b_l = outputs

            if len(x_0.shape) == 2:
                x_0_reshaped = x_0[:, :, None]
            else:
                x_0_reshaped = K.permute_dimensions(x_0, (0, 2, 1))
            x_fake = K.sum(x_0_reshaped, 1)[:, None]
            x_0_reshaped = K.concatenate([x_0_reshaped, x_fake, x_fake], 1)  # (None, n_in+2, n_comp)

            w_b_u = K.concatenate([w_u, b_u[:, None]], 1)  # (None, n_in+1, n_out)
            w_b_l = K.concatenate([w_l, b_l[:, None]], 1)

            w_b = K.concatenate([w_b_u, w_b_l], -1)  # (None, n_in+1, 2*n_out)

            u_l = K.concatenate([u_c, l_c], 1)[:, None]  # (None, 1, 2*n_out)

            w_b_u_l = K.concatenate([w_b, u_l], 1)  # (None, n_in+2, 2*n_out)

            return K.concatenate([x_0_reshaped, w_b_u_l], -1)  # (None, n_in+1, n_comp+2*n_out)

        new_output = Lambda(func)(outputs)

    else:
        raise ValueError(f"Unknown mode {mode}")

    return DecomonModel(
        inputs=inputs,
        outputs=new_output,
        perturbation_domain=model.perturbation_domain,
        ibp=ibp,
        affine=affine,
        finetune=model.finetune,
    )


def get_upper_loss(model: DecomonModel) -> Callable[[tf.Tensor, tf.Tensor], tf.Tensor]:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    n_comp = perturbation_domain.get_nb_x_components()
    n_out = np.prod(model.output[-1].shape[1:])

    def upper_ibp(u_c: tf.Tensor, u_ref: tf.Tensor) -> tf.Tensor:
        # minimize the upper bound compared to the reference
        return K.max(u_c - u_ref, -1)

    def upper_affine(x: tf.Tensor, w_u: tf.Tensor, b_u: tf.Tensor, u_ref: tf.Tensor) -> tf.Tensor:
        upper = perturbation_domain.get_upper(x, w_u, b_u)

        return K.max(upper - u_ref, -1)

    def loss_upper(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        if mode == ForwardMode.IBP:
            u_c = y_pred[:, :, 0]

        elif mode == ForwardMode.AFFINE:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-1, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-1, 0]

            w_u = y_pred[:, :-1, n_comp : n_comp + n_out]
            b_u = y_pred[:, -1, n_comp : n_comp + n_out]

        elif mode == ForwardMode.HYBRID:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-2, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-2, 0]

            w_u = y_pred[:, :-2, n_comp : n_comp + n_out]
            b_u = y_pred[:, -2, n_comp : n_comp + n_out]
            u_c = y_pred[:, -1, n_comp : n_comp + n_out]

        else:
            raise ValueError(f"Unknown mode {mode}")

        if ibp:
            score_ibp = upper_ibp(u_c, y_true)
        if affine:
            score_affine = upper_affine(x_0, w_u, b_u, y_true)

        if mode == ForwardMode.IBP:
            return K.mean(score_ibp)
        elif mode == ForwardMode.AFFINE:
            return K.mean(score_affine)
        elif mode == ForwardMode.HYBRID:
            return K.mean(K.minimum(score_ibp, score_affine))

        raise NotImplementedError()

    return loss_upper


def get_lower_loss(model: DecomonModel) -> Callable[[tf.Tensor, tf.Tensor], tf.Tensor]:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    n_comp = perturbation_domain.get_nb_x_components()
    n_out = np.prod(model.output[-1].shape[1:])

    def lower_ibp(l_c: tf.Tensor, l_ref: tf.Tensor) -> tf.Tensor:
        # minimize the upper bound compared to the reference
        return K.max(l_ref - l_c, -1)

    def lower_affine(x: tf.Tensor, w_l: tf.Tensor, b_l: tf.Tensor, l_ref: tf.Tensor) -> tf.Tensor:
        lower = perturbation_domain.get_lower(x, w_l, b_l)

        return K.max(l_ref - lower, -1)

    def loss_lower(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        if mode == ForwardMode.IBP:
            l_c = y_pred[:, :, 1]

        elif mode == ForwardMode.AFFINE:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-1, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-1, 0]

            w_l = y_pred[:, :-1, n_comp + n_out :]
            b_l = y_pred[:, -1, n_comp + n_out :]

        elif mode == ForwardMode.HYBRID:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-2, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-2, 0]

            w_l = y_pred[:, :-2, n_comp + n_out :]
            b_l = y_pred[:, -2, n_comp + n_out :]
            l_c = y_pred[:, -1, n_comp + n_out :]

        else:
            raise ValueError(f"Unknown mode {mode}")

        if ibp:
            score_ibp = lower_ibp(l_c, y_true)
        if affine:
            score_affine = lower_affine(x_0, w_l, b_l, y_true)

        if mode == ForwardMode.IBP:
            return K.mean(score_ibp)
        elif mode == ForwardMode.AFFINE:
            return K.mean(score_affine)
        elif mode == ForwardMode.HYBRID:
            return K.mean(K.minimum(score_ibp, score_affine))

        raise NotImplementedError()

    return loss_lower


def get_adv_loss(
    model: DecomonModel, sigmoid: bool = False, clip_value: Optional[float] = None, softmax: bool = False
) -> Callable[[tf.Tensor, tf.Tensor], tf.Tensor]:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    n_comp = perturbation_domain.get_nb_x_components()
    n_out = np.prod(model.output[-1].shape[1:])

    def adv_ibp(u_c: tf.Tensor, l_c: tf.Tensor, y_tensor: tf.Tensor) -> tf.Tensor:
        t_tensor = 1 - y_tensor
        s_tensor = y_tensor

        t_tensor = t_tensor[:, :, None]
        s_tensor = s_tensor[:, None, :]
        M = t_tensor * s_tensor
        upper = K.expand_dims(u_c, -1) - K.expand_dims(l_c, 1)
        const = (K.max(upper, (-1, -2)) - K.min(upper, (-1, -2)))[:, None, None]
        upper = upper - (const + K.cast(1, const.dtype)) * (1 - M)
        return K.max(upper, (-1, -2))

    def adv_affine(
        x: tf.Tensor, w_u: tf.Tensor, b_u: tf.Tensor, w_l: tf.Tensor, b_l: tf.Tensor, y_tensor: tf.Tensor
    ) -> tf.Tensor:
        w_u_reshaped = K.expand_dims(w_u, -1)
        w_l_reshaped = K.expand_dims(w_l, -2)

        w_adv = w_u_reshaped - w_l_reshaped
        b_adv = K.expand_dims(b_u, -1) - K.expand_dims(b_l, 1)

        upper = perturbation_domain.get_upper(x, w_adv, b_adv)

        t_tensor = 1 - y_tensor
        s_tensor = y_tensor

        t_tensor = t_tensor[:, :, None]
        s_tensor = s_tensor[:, None, :]
        M = t_tensor * s_tensor

        const = (K.max(upper, (-1, -2)) - K.min(upper, (-1, -2)))[:, None, None]
        upper = upper - (const + K.cast(1, const.dtype)) * (1 - M)
        return K.max(upper, (-1, -2))

    def loss_adv(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        if mode == ForwardMode.IBP:
            u_c = y_pred[:, :, 0]
            l_c = y_pred[:, :, 1]

            if softmax:
                u_c, l_c = softmax_([u_c, l_c], mode=mode, perturbation_domain=perturbation_domain, clip=False)

        elif mode == ForwardMode.AFFINE:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-1, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-1, 0]

            w_u = y_pred[:, :-1, n_comp : n_comp + n_out]
            b_u = y_pred[:, -1, n_comp : n_comp + n_out]
            w_l = y_pred[:, :-1, n_comp + n_out :]
            b_l = y_pred[:, -1, n_comp + n_out :]

            if softmax:
                _, w_u, b_u, w_l, b_l = softmax_(
                    [x_0, w_u, b_u, w_l, b_l], mode=mode, perturbation_domain=perturbation_domain, clip=False
                )

        elif mode == ForwardMode.HYBRID:
            if len(y_pred.shape) == 3:
                x_0 = K.permute_dimensions(y_pred[:, :-2, :n_comp], (0, 2, 1))
            else:
                x_0 = y_pred[:, :-2, 0]

            w_u = y_pred[:, :-2, n_comp : n_comp + n_out]
            b_u = y_pred[:, -2, n_comp : n_comp + n_out]
            w_l = y_pred[:, :-2, n_comp + n_out :]
            b_l = y_pred[:, -2, n_comp + n_out :]
            u_c = y_pred[:, -1, n_comp : n_comp + n_out]
            l_c = y_pred[:, -1, n_comp + n_out :]

            _, u_c, w_u, b_u, l_c, w_l, b_l = softmax_(
                [x_0, u_c, w_u, b_u, l_c, w_l, b_l], mode=mode, perturbation_domain=perturbation_domain, clip=False
            )

        else:
            raise ValueError(f"Unknown mode {mode}")

        if ibp:
            score_ibp = adv_ibp(u_c, l_c, y_true)
            if clip_value is not None:
                score_ibp = K.maximum(score_ibp, clip_value)
        if affine:
            score_affine = adv_affine(x_0, w_u, b_u, w_l, b_l, y_true)
            if clip_value is not None:
                score_affine = K.maximum(score_affine, clip_value)

        if mode == ForwardMode.IBP:
            if sigmoid:
                return K.mean(K.sigmoid(score_ibp))
            else:
                return K.mean(score_ibp)
        elif mode == ForwardMode.AFFINE:
            if sigmoid:
                return K.mean(K.sigmoid(score_affine))
            else:
                return K.mean(score_affine)
        elif mode == ForwardMode.HYBRID:
            if sigmoid:
                return K.mean(K.sigmoid(K.minimum(score_ibp, score_affine)))
            else:
                return K.mean(K.minimum(score_ibp, score_affine))

        raise NotImplementedError()

    return loss_adv


# create a layer
class DecomonLossFusion(DecomonLayer):
    original_keras_layer_class = Layer

    def __init__(
        self,
        asymptotic: bool = False,
        backward: bool = False,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )
        self.convert2mode_layer = Convert2Mode(
            mode_from=mode,
            mode_to=ForwardMode.IBP,
            perturbation_domain=self.perturbation_domain,
        )
        self.asymptotic = asymptotic
        self.backward = backward

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "asymptotic": self.asymptotic,
                "backward": self.backward,
            }
        )
        return config

    def call_no_backward(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if not self.asymptotic:
            u_c, l_c = self.convert2mode_layer(inputs)

            return -l_c + K.log(K.sum(K.exp(u_c - K.max(u_c, -1)[:, None]), -1))[:, None] + K.max(u_c, -1)[:, None]

        else:
            u_c, l_c = self.convert2mode_layer(inputs)
            shape = u_c.shape[-1]

            def adv_ibp(u_c: tf.Tensor, l_c: tf.Tensor, y_tensor: tf.Tensor) -> tf.Tensor:
                t_tensor = 1 - y_tensor
                s_tensor = y_tensor

                t_tensor = t_tensor[:, :, None]
                s_tensor = s_tensor[:, None, :]
                M = t_tensor * s_tensor
                upper = K.expand_dims(u_c, -1) - K.expand_dims(l_c, 1)
                const = (K.max(upper, (-1, -2)) - K.min(upper, (-1, -2)))[:, None, None]
                upper = upper - (const + K.cast(1, const.dtype)) * (1 - M)
                return K.max(upper, (-1, -2))

            source_tensor = tf.linalg.diag(K.ones_like(l_c))

            score = K.concatenate([adv_ibp(u_c, l_c, source_tensor[:, i])[:, None] for i in range(shape)], -1)
            return K.maximum(score, -1)  # + 1e-3*K.maximum(K.max(K.abs(u_c), -1)[:,None], K.abs(l_c))

    def call_backward(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if not self.asymptotic:
            u_c, l_c = self.convert2mode_layer(inputs)
            return K.softmax(u_c)

        else:
            raise NotImplementedError()

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if self.backward:
            return self.call_backward(inputs, **kwargs)
        else:
            return self.call_no_backward(inputs, **kwargs)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[-1]

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        return None


# new layer for new loss functions
class DecomonRadiusRobust(DecomonLayer):
    original_keras_layer_class = Layer

    def __init__(
        self,
        backward: bool = False,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

        if self.mode == ForwardMode.IBP:
            raise NotImplementedError

        if not isinstance(self.perturbation_domain, BoxDomain):
            raise NotImplementedError()

        self.backward = backward

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "backward": self.backward,
            }
        )
        return config

    def call_no_backward(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if self.mode == ForwardMode.HYBRID:
            x, _, w_u, b_u, _, w_l, b_l = inputs
        else:
            x, w_u, b_u, w_l, b_l = inputs

        # compute center
        x_0 = (x[:, 0] + x[:, 1]) / 2.0
        radius = K.maximum((x[:, 1] - x[:, 0]) / 2.0, K.epsilon())
        source_tensor = tf.linalg.diag(K.ones_like(b_l))

        shape = b_l.shape[-1]

        def radius_label(y_tensor: tf.Tensor, backward: bool = False) -> tf.Tensor:
            t_tensor = 1 - y_tensor
            s_tensor = y_tensor

            W_adv = (
                K.sum(-w_l * (s_tensor[:, None]), -1, keepdims=True) + w_u * t_tensor[:, None] + w_l * y_tensor[:, None]
            )  # (None, n_in, n_out)
            b_adv = K.sum(-b_l * s_tensor, -1, keepdims=True) + b_u * t_tensor + (b_l - 1e6) * y_tensor  # (None, n_out)

            score = K.sum(W_adv * x_0[:, :, None], 1) + b_adv  # (None, n_out)

            denum = K.maximum(K.sum(K.abs(W_adv * radius[:, :, None]), 1), K.epsilon())  # (None, n_out)

            eps_adv = K.minimum(-score / denum + y_tensor, 2.0)

            adv_volume = 1.0 - eps_adv

            return K.max(adv_volume, -1, keepdims=True)

        return K.concatenate([radius_label(source_tensor[:, i]) for i in range(shape)], -1)

    def call_backward(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if self.mode == ForwardMode.HYBRID:
            x, _, w_u, b_u, _, w_l, b_l = inputs
        else:
            x, w_u, b_u, w_l, b_l = inputs

        # compute center
        x_0 = (x[:, 0] + x[:, 1]) / 2.0
        radius = K.maximum((x[:, 1] - x[:, 0]) / 2.0, K.epsilon())
        source_tensor = tf.linalg.diag(K.ones_like(b_l))

        shape = b_l.shape[-1]

        def radius_label(y_tensor: tf.Tensor) -> tf.Tensor:
            W_adv = w_u
            b_adv = b_u - 1e6 * y_tensor

            score = K.sum(W_adv * x_0[:, :, None], 1) + b_adv  # (None, n_out)
            denum = K.maximum(K.sum(K.abs(W_adv * radius[:, :, None]), 1), K.epsilon())  # (None, n_out)

            eps_adv = K.minimum(-score / denum + y_tensor, 2.0)

            adv_volume = 1.0 - eps_adv

            return K.max(adv_volume, -1, keepdims=True)

        return K.concatenate([radius_label(source_tensor[:, i]) for i in range(shape)], -1)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        if self.backward:
            return self.call_backward(inputs, **kwargs)
        else:
            return self.call_no_backward(inputs, **kwargs)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[-1]

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        return None


def build_radius_robust_model(model: DecomonModel) -> DecomonModel:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    inputs = model.input
    output = model.output

    layer_robust = DecomonRadiusRobust(
        mode=mode, perturbation_domain=perturbation_domain, backward=model.backward_bounds
    )
    output_robust = layer_robust(output)

    return DecomonModel(
        inputs=inputs,
        outputs=output_robust,
        perturbation_domain=model.perturbation_domain,
        ibp=ibp,
        affine=affine,
        finetune=model.finetune,
    )


##### DESIGN LOSS FUNCTIONS
def build_crossentropy_model(model: DecomonModel) -> DecomonModel:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    inputs = model.input
    output = model.output

    layer_fusion = DecomonLossFusion(mode=mode, backward=model.backward_bounds)
    output_fusion = layer_fusion(output, mode=mode)

    return DecomonModel(
        inputs=inputs,
        outputs=output_fusion,
        perturbation_domain=model.perturbation_domain,
        ibp=ibp,
        affine=affine,
        finetune=model.finetune,
    )


##### DESIGN LOSS FUNCTIONS
def build_asymptotic_crossentropy_model(model: DecomonModel) -> DecomonModel:
    ibp = model.ibp
    affine = model.affine

    mode = get_mode(ibp, affine)
    perturbation_domain = model.perturbation_domain

    inputs = model.input
    output = model.output

    layer_fusion = DecomonLossFusion(mode=mode, asymptotic=True)
    output_fusion = layer_fusion(output, mode=mode)

    return DecomonModel(
        inputs=inputs,
        outputs=output_fusion,
        perturbation_domain=model.perturbation_domain,
        ibp=ibp,
        affine=affine,
        finetune=model.finetune,
    )



import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer, Reshape
from tensorflow.python.keras.backend import _get_available_gpus

from decomon.backward_layers.convert import to_backward
from decomon.core import ForwardMode, Slope
from decomon.layers.decomon_layers import DecomonActivation, DecomonFlatten
from decomon.layers.decomon_reshape import DecomonReshape


def test_Backward_Activation_1D_box_model(n, activation, mode, floatx, decimal, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # decomon layer
    decomon_layer = DecomonActivation(activation, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(decomon_layer)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_Activation_1D_box_model_slope(helpers):
    n = 2
    activation = "relu"
    mode = ForwardMode.AFFINE
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    #  decomon layer
    decomon_layer = DecomonActivation(activation, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # to_backward with a given slope => outputs
    outputs_by_slope = {}
    for slope in Slope:
        layer_backward = to_backward(decomon_layer, slope=slope, mode=mode)
        assert layer_backward.slope == slope
        outputs = layer_backward(inputs_for_mode)
        f_decomon = K.function(inputs, outputs)
        outputs_by_slope[slope] = f_decomon(inputs_)

    # check results
    # O_Slope != Z_Slope
    same_outputs_O_n_Z = [
        (a == b).all() for a, b in zip(outputs_by_slope[Slope.O_SLOPE], outputs_by_slope[Slope.Z_SLOPE])
    ]
    assert not all(same_outputs_O_n_Z)

    # V_Slope == Z_Slope
    for a, b in zip(outputs_by_slope[Slope.V_SLOPE], outputs_by_slope[Slope.Z_SLOPE]):
        assert (a == b).all()


def test_Backward_Activation_multiD_box(odd, activation, floatx, decimal, mode, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    #  decomon layer
    decomon_layer = DecomonActivation(activation, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(decomon_layer)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_Flatten_multiD_box(odd, floatx, decimal, mode, data_format, helpers):
    if data_format == "channels_first" and not len(_get_available_gpus()):
        pytest.skip("data format 'channels first' is possible only in GPU mode")

    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    #  decomon layer
    decomon_layer = DecomonFlatten(data_format, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(decomon_layer)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_Reshape_multiD_box(odd, floatx, decimal, mode, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    #  decomon layer
    decomon_layer = DecomonReshape((-1,), dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # get backward layer
    backward_layer = to_backward(decomon_layer)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_to_backward_ko():
    class MyLayer(Layer):
        ...

    layer = MyLayer()
    with pytest.raises(NotImplementedError):
        to_backward(layer)


def test_name():
    layer = Reshape((1, 2))
    backward_layer = to_backward(layer)
    backward_layer.name.startswith(f"{layer.name}_")



import tensorflow.python.keras.backend as K

from decomon.core import ForwardMode, Slope
from decomon.layers.decomon_layers import DecomonActivation


def test_decomon_activation_slope(helpers):
    mode = ForwardMode.AFFINE
    activation = "relu"
    n = 2

    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=False)
    (
        x_0,
        y_0,
        z_0,
        u_c_0,
        W_u_0,
        b_u_0,
        l_c_0,
        W_l_0,
        b_l_0,
    ) = inputs_  # numpy values

    outputs_by_slope = {}
    for slope in Slope:
        layer = DecomonActivation(activation, dc_decomp=False, mode=mode, slope=slope)
        assert layer.slope == slope
        inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=False)
        x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs  # tensors
        inputs_for_mode = [z, W_u, b_u, W_l, b_l]
        output = layer(inputs_for_mode)
        f_func = K.function(inputs, output)
        outputs_by_slope[slope] = f_func(inputs_)

    # check results
    # O_Slope != Z_Slope
    same_outputs_O_n_Z = [
        (a == b).all() for a, b in zip(outputs_by_slope[Slope.O_SLOPE], outputs_by_slope[Slope.Z_SLOPE])
    ]
    assert not all(same_outputs_O_n_Z)

    # V_Slope == Z_Slope
    for a, b in zip(outputs_by_slope[Slope.V_SLOPE], outputs_by_slope[Slope.Z_SLOPE]):
        assert (a == b).all()



import pytest
import tensorflow as tf
from tensorflow.keras.layers import Dense

from decomon.keras_utils import get_weight_index_from_name


def test_get_weight_index_from_name_nok_attribute():
    layer = Dense(3, input_dim=1)
    layer(tf.zeros((2, 1)))
    with pytest.raises(AttributeError):
        get_weight_index_from_name(layer=layer, weight_name="toto")


def test_get_weight_index_from_name_nok_index():
    layer = Dense(3, input_dim=1, use_bias=False)
    layer(tf.zeros((2, 1)))
    with pytest.raises(IndexError):
        get_weight_index_from_name(layer=layer, weight_name="bias")


def test_get_weight_index_from_name_ok():
    layer = Dense(3, input_dim=1)
    layer(tf.zeros((2, 1)))
    assert get_weight_index_from_name(layer=layer, weight_name="bias") in [0, 1]



#  Copyright (c) 2022 AIRBUS and its affiliates.
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this source tree.

import importlib
import logging
import pkgutil
import sys

import decomon

logger = logging.getLogger(__name__)


def find_abs_modules(package):
    """Find names of all submodules in the package

    https://stackoverflow.com/questions/48879353/how-do-you-recursively-get-all-submodules-in-a-python-package

    """
    path_list = []
    spec_list = []
    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__):
        import_path = f"{package.__name__}.{modname}"
        if ispkg:
            spec = pkgutil._get_spec(importer, modname)
            try:
                importlib._bootstrap._load(spec)
                spec_list.append(spec)
            except Exception as e:
                logger.warning(f"Could not load package {modname}, so it will be ignored ({e}).")
        else:
            path_list.append(import_path)
    for spec in spec_list:
        del sys.modules[spec.name]
    return path_list


def test_importing_all_submodules():
    modules_with_errors = []
    for m in find_abs_modules(decomon):
        try:
            importlib.import_module(m)
        except Exception as e:
            modules_with_errors.append(m)
            print(f"{m}: {e.__class__.__name__}: {e}")
    if len(modules_with_errors) > 0:
        raise ImportError(
            f"{len(modules_with_errors)} submodules of decomon cannot be imported\n" + f"{modules_with_errors}"
        )


if __name__ == "__main__":
    test_importing_all_submodules()
